/*
** avfilter.c - written in milano by vesely on 19 oct 2001
** sophos anti virus for courier-mta global filters
*/
/*
** Copyright (c) 1999, 2000, 2001 Alessandro Vesely
** All rights reserved. see NOTICE.txt
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <syslog.h>
#include <courier/filtersocketdir.h>
 
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <signal.h>
#include <courier/courier.h> /* for config_localfilename */
#include <libfilter/libfilter.h>

/* SAV */
#include "sav_if/compute.h"
/*
 * Depending upon your platform, your compiler, and whether or
 * not you are linking with the SAVI library statically or
 * dynamically, you may need to define INITGUID before
 * including csavi2c.h.
 * You can do this either by changing the source at this point,
 * or by instructing your compiler (usually -DINITGUID).
 *
 * -ale: INITGUID is a declare/define switch that works in swiid
 * (also included from csavi2c.h)
 */
#define INITGUID
#include "sav_if/csavi2c.h"
#undef INITGUID
#include "sav_if/swiid.h"

/* MIMEC */
#include "mime.h"
#include "parse.h"

int signal_timed_out = 0,
	signal_break = 0,
	signal_hangup = 0,
	verbose = 0,
	live_children = 0;

int keep_running(void)
{
	return signal_timed_out == 0 && signal_break == 0;
}

void child_reaper(int sig)
{
	int status;
	pid_t child;

	(void)sig;
	
	while ((child = waitpid(-1, &status, WNOHANG)) > 0)
	{
		--live_children;
#if !defined(NDEBUG)
		fprintf(stderr, "avfilter: %d exited, %d remaining\n",
			(int)child, live_children);
#endif
	}
}

void sig_catcher(int sig)
{
#if !defined(NDEBUG)
	fprintf(stderr, "avfilter: %d received signal %d: keep=%d\n",
		(int)getpid(), sig, keep_running());
#endif
	switch(sig)
	{
		case SIGALRM:
			signal_timed_out = 1;
			break;

		case SIGHUP:
			signal_hangup = 1;
			break;
		
		case SIGPIPE:
		case SIGINT:
		case SIGQUIT:
		case SIGTERM:
			signal_break = 1;
			break;
		default:
			break;
	}
}

void set_alarm(unsigned seconds)
{
#if !defined(NDEBUG)
	struct sigaction oact;
	if (sigaction(SIGALRM, NULL, &oact) != 0 ||
		oact.sa_handler != &sig_catcher)
			fprintf(stderr, "SIGALRM not correct!!\n");
#endif	
	signal_timed_out = 0;
	alarm(seconds);
}

static void save_file(FILE *fp, char const *dir, char const *msg)
{
	char templ[] = "avf-XXXXXXX";
	unsigned namesize = strlen(dir) + sizeof templ + 4;
	char *fname = (char*)malloc(namesize); /* tempnam(dir, "avf-"); */
	char *error;
	int fno;
	
	if (fname == NULL) error = "malloc fail";
	else if (namesize >= PATH_MAX) error = "path too long";
	else
	{
		sprintf(fname, "%s/%s", dir, templ);
		fno = mkstemp(fname);
		if (fno == -1)
			error = "mkstemp fails";
		else
			error = NULL;
	}
	
	if (error)
	{
		fprintf(stderr, "avfilter: Unable to save %.*s: %s\n",
			PATH_MAX, fname ? fname: "(null)", error);
	}
	else
	{
		FILE *fps = fdopen(fno, "w+");
		char buf[4096];
		size_t r = 1;
		if (fps)
		{
			fseek(fp, 0, SEEK_SET);
			while ((r = fread(buf, 1, sizeof buf, fp)) > 0)
				if (fwrite(buf, r, 1, fps) != 1)
					break;
			r = (ferror(fp) || ferror(fps)) ? 1 : 0;
			fclose(fps);
		}
		else
			close(fno);
		if (r)
		{
			fprintf(stderr, "avfilter: Unable to save %s: %s\n",
				fname, strerror(errno));
			unlink(fname);
		}
		else
			fprintf(stderr, "avfilter: Saved %s file in %s\n", msg, fname);
		free(fname);
	}
}

static int is_discard_header(char const *head, size_t size)
{
	static char const * const discard_header[] = { "to:", "cc", "bcc" };
	unsigned i;

	for (i = 0; i < sizeof discard_header / sizeof discard_header[0]; ++i)
	{
		char const * const p = discard_header[i];
		unsigned len = strlen(p);
		if (len < size && strincmp(head, p, len) == 0)
			return len;
	}
	return 0;
}

static int read_message_header(FILE *fp, char *buf, size_t size)
/*
** return -2 = i/o error, -1 = length error,
** 0 = end of headers, len = length of header read
*/
{
	int flag = 1;
	unsigned len = 0;
	while (len < size && fgets(&buf[len], size - len, fp) != NULL)
	{
		int ch;
		
		len = strlen(buf);
		if (len <= 0 || buf[len - 1] != '\n')
			return -1;
		if (len == 1)
			return 0; /* end of headers */
		
		/* check for wrapped header */
		ch = fgetc(fp);
		if (ch == '\n' || !isspace(ch))
		{
			ungetc(ch, fp);
			buf[--len] = 0;
			return len;
		}
		buf[len++] = ch;
		
		if (flag)
		{
			if (flag == 1 &&
				(flag = is_discard_header(buf, len)) == 0)
					continue;
			/*
			** discard all but last line of MIME-irrelevant headers (to, cc)
			*/
			len = flag;
		}
	}
	return len >= size ? -1 : -2;
}

typedef struct virus_found_msg_struct
{
	char *errmsg, *lasterrmsg;
	size_t errmsglen;
} virus_msg;

static void virus_msg_init(virus_msg *v)
{
	memset(v, 0, sizeof *v);
}

static int virus_msg_add(virus_msg *v,
	char const *name, char const *location, int virus)
{
	static char const fmt[] = "554 %s \"%s\" in %s\n";
	size_t required = strlen(name) + strlen(location) + (sizeof fmt + 5);
	if (v->lasterrmsg)
		v->lasterrmsg[3] = '-';
	if ((v->lasterrmsg = (char*)malloc(required + v->errmsglen)) == NULL)
		return 1;
	if (v->errmsg)
	{
		memcpy(v->lasterrmsg, v->errmsg, v->errmsglen);
		free(v->errmsg);
	}
	v->errmsg = v->lasterrmsg;
	v->lasterrmsg += v->errmsglen;
	v->errmsglen += sprintf(v->lasterrmsg, fmt,
		virus ? "Virus" : "Error", name, location);
	return 0;
}

static void virus_msg_clear(virus_msg *v)
{
	if (v->errmsg)
		free(v->errmsg);
}

static int ListInfections(CIEnumSweepResults *pEnumResults,
	virus_msg *vm, char const *location)
{
  const U32       BufLen     = 80;
  CISweepResults *pResults   = NULL;
  HRESULT         hr;
  OLECHAR         Buf[81];
  SOPHOS_ULONG    pcFetched;
  
  /*
   * Sanity.
   */
  if ((!location) || (!vm) || (!pEnumResults))
  {
    fprintf(stderr,
			"avfilter: ListInfections() called with incorrect parameters.\n");
    return 1;
  }

  /*
   * Ensure that the enumerator is positioned
   * at the start.  If this fails, return now.
   */
  hr = pEnumResults->pVtbl->Reset(pEnumResults);
  if (SOPHOS_FAILED(hr))
  {
    fprintf(stderr, "avfilter: Failed to reset enumerator [0x%08lx].\n",
    	(unsigned long)hr);
    return 1;
  }

  /*
   * Walk through the enumeration interface,
   * retrieving one results object at a time.
   */
	while (pEnumResults->pVtbl->Next(pEnumResults,
		1, (void **)&pResults, &pcFetched) == SOPHOS_S_OK)
	{
		hr = pResults->pVtbl->GetVirusName(pResults, BufLen, (LPOLESTR) Buf, NULL);
		if (SOPHOS_FAILED(hr))
		{
			fprintf(stderr, "avfilter: Failed to get virus name [0x%08lx].\n",
				(unsigned long)hr);
			strcpy(Buf, "Unknown");
		}
		
		if (virus_msg_add(vm, Buf, location, 1))
			goto bail;
		
		pResults->pVtbl->Release(pResults);
		pResults = NULL;
  }
  
  return 0;

bail:
  if (pResults)
    pResults->pVtbl->Release(pResults);
  return 1;
}

static char *read_fname(int fd)
{
	char buf[1024], *rtc = NULL;
	unsigned l = 0, count = 0;
	int found = 0, empty = 0;

#if !defined(NDEBUG)
	fprintf(stderr, "avfilter: reading fd %d from proc. %d\n",
		fd, (int)getpid());
#endif
	
	set_alarm(30);
	while (l < sizeof buf && empty == 0 && keep_running())
	{
		unsigned i = l;
		unsigned const p = read(fd, &buf[l],
			1 /* sizeof buf - l hangs reading unavailable data */);
		if (p == (unsigned)(-1))
		{
			switch (errno)
			{
				case EAGAIN:
				case EINTR:
					continue;
				default:
					break;
			}
			fprintf(stderr, "avfilter: cannot read ctlfile: %s\n",
				strerror(errno));
			break;
		}
		else if (p == 0)
		{
			fprintf(stderr, "avfilter: Unexpected EOF on ctlfile\n");
			break;
		}
		l += p;
		while (i < l /* useless: && keep_running() */)
		{
			if (buf[i] != '\n')
				++count;
			else if (count == 0) /* empty line */
			{
				buf[i] = 0;
				l = i;
				empty = 1;
			}
			else 
			{
				count = 0;
				if (found == 0) /* first line */
				{
					found = 1;
					buf[i] = '&';
					rtc = (char*)malloc(i+1);
					if (rtc)
					{
						memcpy(rtc, buf, i);
						rtc[i] = 0;
					}
				}
				else
					buf[i] = ';';
			}
			++i;
		}
	}
	
	alarm(0);
	if (!keep_running())
		fprintf(stderr, "avfilter: reading ctlfile terminated prematurely\n");
	if (rtc == NULL)
		fprintf(stderr, "avfilter: No file name\n");
	if (l >= sizeof buf)
		fprintf(stderr, "avfilter: Buffer overflow reading ctlfile\n");
#if !defined(NDEBUG)		
	fprintf(stderr, "avfilter: ctlfile content: %.*s\n",
		(int)l, buf);
#endif		
	
	return rtc;
}

/*
 * Function: InitialiseSAVI
 *
 * Parameters: (1) pointer to the SAVI client name.
 *
 * Returns: Initialised SAVI interface ready for use
 *          on success.
 *          NULL on failure.
 *
 * Purpose: Performs the bootstrap process -- creates
 *          the SAVI interface from scratch, and
 *          initialises it ready for use.
 */
static CISavi2 *InitialiseSAVI(LPCOLESTR ClientName)
{
  CISavi2              *pSAVI = NULL;
  CISweepClassFactory2 *pFactory =NULL;
  HRESULT               hr;
  U32               vdataVerStringLength  = 100;
  CIEnumIDEDetails *IDElist       = NULL;
  OLECHAR           vdataVerString[101];
  SYSTEMTIME        date;
  U32               virNum;
  U32               vengVer;
    
  /*
   * Request a class factory interface.
   */
  hr = DllGetClassObject((REFIID)&SOPHOS_CLSID_SAVI2,
                         (REFIID)&SOPHOS_IID_CLASSFACTORY2,
                         (void **)&pFactory);
  if (SOPHOS_FAILED(hr))                                                                          
  {
    fprintf(stderr, "ERROR: Failed to get class factory interface [0x%08lx].\n",
    	(unsigned long)hr);
    return NULL;
  }

  /*
   * Ask the class factory for a CSAVI2 interface.
   */
  hr = pFactory->pVtbl->CreateInstance(pFactory,
                                       NULL,
                                       &SOPHOS_IID_SAVI2,
                                       (void **)&pSAVI);

  /*
   * Drop the factory immediately, we don't need it
   * again in this example.
   */
  pFactory->pVtbl->Release(pFactory);

  /*
   * Did we get the CSAVI2 interface we requested?
   */
  if (SOPHOS_FAILED(hr))
  {
    fprintf(stderr, "ERROR: Failed to get a CSAVI2 interface [0x%08lx].\n",
    	(unsigned long)hr);
    return NULL;
  }

  /*
   * Ask SAVI to initialise itself.
   */
	hr = pSAVI->pVtbl->InitialiseWithMoniker(pSAVI, ClientName);

  /*
   * If the initialisation failed, then release the SAVI
   * interface and set the pointer to NULL.
   */
  if (SOPHOS_FAILED(hr))
  {
    fprintf(stderr, "ERROR: Failed to initialise SAVI [0x%08lx].",
    	(unsigned long)hr);
    pSAVI->pVtbl->Release(pSAVI);
    pSAVI = NULL;
  }
  else
  {
    /*
     * Ask for the engine information - virus engine and data version number,
     * and the number of viruses.
     */
    hr = pSAVI->pVtbl->GetVirusEngineVersion(pSAVI,
                                             &vengVer,
                                             (LPOLESTR) vdataVerString,
                                             vdataVerStringLength,
                                             &date,
                                             &virNum,
                                             NULL,
                                             (REFIID)&SOPHOS_IID_ENUM_IDEDETAILS,
                                             (void **)&IDElist);

    if (SOPHOS_FAILED(hr))
    {
      fprintf(stderr, "avfilter: Error getting version number from engine\n");
    }
    else
    {
      fprintf(stderr, "avfilter: Virus engine version: %u.%u\n", 
                     (unsigned int) ((vengVer & 0xFFFF0000) >> 16),
                     (unsigned int) (vengVer & 0x0000FFFF) );
      fprintf(stderr, "avfilter: Virus data version:   %s\n", vdataVerString);
      fprintf(stderr, "avfilter: Virus data date:      %d/%d/%d\n",
      	date.wDay, date.wMonth, date.wYear);
      fprintf(stderr, "avfilter: No. of viruses:       %ld\n", virNum);
    }
  }

  /*
   * Return the SAVI interface.  May be NULL if creation
   * or initialisation failed.
   */
  return pSAVI;
}

typedef struct savi_scan_error_struct
{
	HRESULT hr;
	char const *msg; /* printed by virus_msg_add(..., 0) */
} savi_se;

typedef struct avfilter_parm_struct
{
	int err; /* error parsing config file */
	char *config_file;
	
	/* strings in PARSE */
	char *save_missed_file;
	char *save_virus_file;
	char *tmp_dir;
	savi_se *savi_se_table;
	unsigned savi_se_table_size;
	PARSE *parse;
	CISavi2 *pSAVI;
} av_parm;

static int parm_init(av_parm *parm)
/*
** initialize parameters, loading AV software
*/
{
	char *fname;
	
	memset(parm, 0, sizeof *parm);
	if ((parm->pSAVI = InitialiseSAVI("avfilter")) == NULL)
		return 1;
	fname = parm->config_file = config_localfilename("filters/avfilter.conf");
	if (fname == NULL)
		parm->err = 1;
	else if ((parm->parse = parse_init(fname)) == NULL)
		parm->err = (errno == ENOENT ? 0 : 1);

	return parm->err;
}

static void parm_clear(av_parm *parm)
{
	CISavi2 *const pSAVI = parm->pSAVI;
	if (pSAVI)
	{
		pSAVI->pVtbl->Terminate(pSAVI);
		pSAVI->pVtbl->Release(pSAVI);
	}
	if (parm->config_file)
		free(parm->config_file);
	if (parm->savi_se_table)
		free(parm->savi_se_table);
	
#if 0
	/* directories kept in parse buffer */
	if (parm->save_virus_file)
		free(parm->save_virus_file);
	if (parm->save_missed_file)
		free(parm->save_missed_file);
	if (parm->tmp_dir)
		free(parm->tmp_dir);
#endif

	if (parm->parse)
		parse_clear(parm->parse);
	memset(parm, 0, sizeof *parm);
}

static char const* read_message(MIME *mime, FILE *fp)
{
	char buf[4096];
	int rtc, mime_headers = 0;
	
	while ((rtc = read_message_header(fp, buf, sizeof buf)) > 0)
	{
		int x = mime_set_header(mime, mime_header_unknown, buf);
		if (x < 0)
		{
			rtc = -2;
			break;
		}
		if (x > 0)
			++mime_headers;
	}
	
	if (rtc == 0 && mime_headers == 0 /* is text/plain by default */
	|| rtc == -2 && ferror(fp) == 0) /* only headers given */
		return "200 Ok.\n"; /* nothing to scan */
	
	if (rtc == 0 && mime_read(mime, fp) == 0)
		return NULL; /* good return */

	switch (rtc)
	{
		case -2: return "451 system failure in avfilter\n";
		case -1: return "554 malformed message headers\n";
		default: return "554 malformed MIME structure\n";
	}
}

static int savi_scan_error(av_parm const *parm, HRESULT hr,
	virus_msg *vm, char const *location)
{
	int i, rtc = 0;
	int const table_size = parm->savi_se_table_size;
	savi_se const* const table = parm->savi_se_table;
	
	for (i = 0; i < table_size; ++i)
	{
		savi_se const * const se = &table[i];
		if (hr == se->hr)
		{
			if (se->msg != NULL)
				rtc = virus_msg_add(vm, se->msg, location, 0);
			break;
		}
	}
	
	if (i >= table_size)
	{
		rtc = virus_msg_add(vm, "cannot scan content", location, 0);
		fprintf(stderr, "avfilter: Cannot scan %s: [0x%08lx].\n",
			location, (unsigned long)hr);
	}
	return rtc;
}

static int scan_mime_part(av_parm const *parm, MIME *mime, virus_msg *vm)
{
	char const*const tmp_dir = parm->tmp_dir ? parm->tmp_dir : "/tmp";
	char templ[] = "avf-XXXXXXX";
	unsigned namesize = strlen(tmp_dir) + sizeof templ + 4;
	char *fname = (char*)malloc(namesize); /* tempnam(tmp_dir, "avf-"); */
	char *error;
	int fno, rtc = 1;
	
	if (fname == NULL) error = "malloc fail";
	else if (namesize >= PATH_MAX) error = "path too long";
	else
	{
		sprintf(fname, "%s/%s", tmp_dir, templ);
		fno = mkstemp(fname);
		if (fno == -1)
			error = "mkstemp fails";
		else
			error = NULL;
	}
	
	if (error)
	{
		fprintf(stderr, "avfilter: Cannot create mime file %.*s: %s\n",
			PATH_MAX, fname ? fname: "(null)", error);
	}
	else
	{
		FILE *fp = fdopen(fno, "w+");
		if (fp)
		{
			int const xtract =
				mime_extract_single_body(mime, fp, mime_binary, 1) == 0;
			fclose(fp);
			if (xtract)
			{
				CIEnumSweepResults *pEnumResults = NULL;
				CISavi2 *const pSAVI = parm->pSAVI;
				HRESULT hr = pSAVI->pVtbl->SweepFile(pSAVI, fname,
					(REFIID)&SOPHOS_IID_ENUM_SWEEPRESULTS,
					(void **)&pEnumResults);
				char locbuf[32];
				char const *location = mime_get_filename(mime);
				if (location == NULL)
				{
					location = mime_get_name(mime);
					if (location == NULL)
					{
						sprintf(locbuf, "part #%d", mime_get_position(mime));
						location = locbuf;
					}
				}
				
				/* set errors (virus or unscannable item) */
				if (SOPHOS_SUCCEEDED(hr))
					rtc = ListInfections(pEnumResults, vm, location);
				else
					rtc = savi_scan_error(parm, hr, vm, location);
				
				if (pEnumResults)
					pEnumResults->pVtbl->Release(pEnumResults);
			}
			else
				fprintf(stderr, "avfilter: Cannot extract MIME part %d: %s (%s)\n",
					mime_get_position(mime),
					mime_get_last_error_string(mime),
					strerror(errno));
		}
		else
		{
			fprintf(stderr, "avfilter: Cannot open %s: %s\n",
				fname, strerror(errno));
			close(fno);
		}
		unlink(fname);
		free(fname);
	}
	return rtc;
}

static char const* avfilter(av_parm *parm, FILE *fp)
{
	static char const*const resp_ok = "200 Ok.\n";
	static char const*const resp_bad = "451 system malfunction in avfilter\n";
	MIME *mime = mime_new(mime_encoding_not_set, mime_charset_not_set);
	char const *rtc = read_message(mime, fp);

	if (rtc == NULL)
	{
		size_t const parts = mime_get_count_all(mime);
		size_t i;
		virus_msg vm;
		virus_msg_init(&vm);
		
		for (i = 0; i < parts; ++i)
		{
			mime_set_position(mime, i);
			if (mime_is_leaf_part(mime) &&
				mime_get_mime_subtype(mime) != mime_text_plain &&
				!mime_is_empty(mime) &&
				scan_mime_part(parm, mime, &vm) != 0)
			{
				rtc = resp_bad;
				break;
			}
		}
		if (rtc == NULL &&
			vm.errmsg != NULL &&
			(rtc = strdup(vm.errmsg)) == NULL) /* not freed until exit */
				rtc = resp_bad;
		
		virus_msg_clear(&vm);
	}
	mime_clear(mime);
	if (rtc == NULL)
		rtc = resp_ok;
	else
	{
		char const ch = *rtc;
		if (ch == '4' && parm->save_missed_file)
			save_file(fp, parm->save_missed_file, "missed");
		if (ch == '5' && parm->save_virus_file)
			save_file(fp, parm->save_virus_file, "virus");
	}
	return rtc;
}

static void runchild(int fd_in, int fd_out, av_parm *parm)
{
	pid_t pid;
	int retry = 2;
	
	while ((pid = fork()) < 0 && --retry >= 0)
		sleep(1);
	
	if (pid == 0) /* child */
	{
		char *fname;
		if (verbose > 1)
			fprintf(stderr, "avfilter[%d]: started child\n", (int)getpid());
#if !defined(NDEBUG)
		{
			char *debug = getenv("DEBUG_AVFILTER");
			if (debug)
			{
				int nsec = atoi(debug);
				if (nsec <= 0) nsec = 20;
				
				/* hallo! */
				fprintf(stderr, "DEBUG_AVFILTER: %s\n"
					"now you have %d secs to\n"
					"%% gdb <path> %d\n"
					"(gdb) list %d\n",
						debug, nsec, (int)getpid(), __LINE__ );
				sleep(nsec);
			}
			else
				fprintf(stderr,
					"set DEBUG_AVFILTER to debug the child in gdb\n");
		}
#endif
		fname = read_fname(fd_in);
		if (fname)
		{
			FILE *fp = fopen(fname, "r+");
			if (fp)
			{
				char const *resp = avfilter(parm, fp);
				unsigned w = 0, l = strlen(resp);
				
				fclose(fp);
				if (verbose)
					fprintf(stderr, "avfilter[%d]: %s%s", (int)getpid(),
						keep_running() ? "" : "interrupted: ", resp);
				
				/* give response
				** only if not interrupted or timed out, 
				*/
				while (w < l && keep_running())
				{
					unsigned p = write(fd_out, resp + w, l - w);
					if (p == (unsigned)(-1))
					{
						switch (errno)
						{
							case EAGAIN:
							case EINTR:
								continue;
							default:
								break;
						}
						fprintf(stderr, "avfilter[%d]: unable to write resp: %s\n",
							(int)getpid(), strerror(errno));
						break;
					}
					w += p;
				}
			}
			else
				fprintf(stderr, "avfilter: cannot open %s: %s\n",
					fname, strerror(errno));
			free(fname);
		}
		exit(0);
	}
	else  if (pid > 0) /* parent */
		++live_children;
	else
		perror("avfilter: fork");
}

static int set_savi_scan_error(av_parm *parm, PARSE *p)
{
	char *err_code = NULL, *msg = NULL;
	int err = 0;
	int tok = get_token(p);
	HRESULT hr;
	
	if (tok == '_')
	{
		err_code = get_word_delimited(p, "_", 0, 1);
		if (err_code)
		{
			tok = get_token(p);
			err |= (tok != '=');
			msg = get_toeol(p);
			tok = get_token(p);
			if (err == 0 && msg != NULL)
			{
				char *t = NULL;
				hr = strtoul(err_code, &t, 0);
				if (t == NULL || *t != 0 || SOPHOS_SUCCEEDED(hr))
					err = 1;
			}
		}
	}
	
	if (err || err_code == NULL || msg == NULL)
	{
		fprintf(stderr,
			"avfilter: invalid error code \"%s\" = \"%s\" at line %d in %s\n",
			Psz(err_code), Psz(msg), p->line, p->fname);
		parm->err = 1;
	}
	else
	{
		int const ndx = parm->savi_se_table_size;
		savi_se *table = (savi_se*)malloc((ndx + 1) * sizeof(savi_se));
		if (table == NULL)
		{
			fprintf(stderr,
				"avfilter: cannot allocate %d savi table elements at line %d in %s:\n",
				ndx, p->line, p->fname);
			parm->err = 1;
		}
		if (parm->savi_se_table != NULL)
		{
			memcpy(table, parm->savi_se_table, ndx * sizeof(savi_se));
			free(parm->savi_se_table);
		}
		table[ndx].hr = hr;
		table[ndx].msg = *msg ? msg : NULL;
		parm->savi_se_table = table;
		parm->savi_se_table_size = ndx + 1;
	}
	
	return tok;
}

static int set_savi_option(av_parm *parm, PARSE *p)
{
	CISavi2 *const pSAVI = parm->pSAVI;
	char *name = NULL, *value = NULL, *type = NULL;
	int vtype = 3, err = 0;
	int tok = get_token(p);
	
	if (tok == '_')
	{
		name = get_word_delimited(p, "-_", 0, 1);
		if (name)
		{
			tok = get_token(p);
			if (tok == '[')
			{
				type = get_word_until(p, ']');
				if (type)
				{
					char *t = NULL;
					long l = strtoul(type, &t, 10);
					if (t && *t == 0 && l < 10)
						vtype = (int)l;
					else
						err = 1;
				}
				if ((tok = get_token(p)) == ']')
					tok = get_token(p);
			}
			err |= (tok != '=');
			value = get_toeol(p);
			tok = get_token(p);
		}
	}
	
	if (err || name == NULL || value == NULL)
	{
		fprintf(stderr, "avfilter: invalid \"%s\" [%s] = \"%s\" at line %d in %s\n",
			Psz(name), Psz(type), Psz(value), p->line, p->fname);
		parm->err = 1;
	}
	else
	{
		HRESULT hr = pSAVI->pVtbl->SetConfigValue(pSAVI, name, vtype, value);
		if (SOPHOS_FAILED(hr))
		{
			fprintf(stderr,
				"avfilter: cannot set %s[%d]=%s at line %d in %s: [0x%08lx]\n",
				name, vtype, value, p->line, p->fname, (unsigned long)hr);
			parm->err = 1;
		}
	}
	
	return tok;
}

int configure_directory(char **dir, av_parm *parm)
{
	PARSE *const p = parm->parse;
	int tok = get_token(p);
	if (tok == '=')
	{
		char *d = get_toeol(p);
		if (d)
		{
			struct stat s;
			errno = 0;
			if (stat(d, &s) != 0 ||
				!S_ISDIR(s.st_mode) ||
				(S_IRWXU & s.st_mode) != S_IRWXU ||
				(S_IRWXO & s.st_mode) != 0 ||
				s.st_uid != MAILUID)
			{
				fprintf(stderr, "avfilter: unsuitable directory %s: %s in %s line %d\n",
					d,
					errno ? strerror(errno) :
						!S_ISDIR(s.st_mode) ? "not a directory" :
						(S_IRWXU & s.st_mode) != S_IRWXU ? "not rwx by " MAILUSER:
						(S_IRWXO & s.st_mode) != 0 ? "world accessible" :
						s.st_uid != MAILUID ? "not owned by " MAILUSER : "uh?",
					p->fname, p->line);
				parm->err = 1;
			}
			else
				*dir = d; /* kept there until parse_clear */
			tok = get_token(p);
		}
		else
		{
			parm->err = 1;
			tok = -1;
		}
	}
	else
	{
		fprintf(stderr,
			"avfilter: '=' followed by directory name expected in %s at line %d\n",
			p->fname, p->line);
		parm->err = 1;
	}
	return tok;
}

static int configure(av_parm *parm)
/*
** read the parameter file and set any option
*/
{
	PARSE *const p = parm->parse;
	int tok;
	char *next = NULL;

	if (p == NULL)
		return parm->err;

#if !defined(NDEBUG)
	fprintf(stderr, "reading parameters from %s\n", p->fname);
#endif
	
	tok = get_token(p);
	while (tok >= 0)
	{
		if (tok == '_')
		{
			next = get_next(p);
			if (next == NULL)
				break;
		}
		else
		{
			if (tok > 0)
				fprintf(stderr, "avfilter: Invalid token %c at line %d in %s\n",
					tok, p->line, p->fname);
			break;
		}
	
		if (strcmp(next, "set") == 0)
			tok = set_savi_option(parm, p);
		else if (strcmp(next, "scan_error") == 0)
			tok = set_savi_scan_error(parm, p);
		else if (strcmp(next, "verbose") == 0)
		{
			verbose = 1;
			if ((tok = get_token(p)) == '=')
			{
				if ((tok = get_token(p)) == '_' &&
					(next = get_next(p)) != NULL)
				{
					char *t = NULL;
					long l = strtoul(next, &t, 10);
					tok = get_token(p);
					if (t && *t == 0 && l < 10)
						verbose = (int)l;
					else
						tok = -1;
				}
				else
					tok = -1;
			}
		}
		else if (strcmp(next, "save_missed") == 0)
			tok = configure_directory(&parm->save_missed_file, parm);
		else if (strcmp(next, "save_virus") == 0)
			tok = configure_directory(&parm->save_virus_file, parm);
		else if (strcmp(next, "tmp_dir") == 0)
			tok = configure_directory(&parm->tmp_dir, parm);
		else
		{
			fprintf(stderr, "invalid verb \"%s\" at line %d in %s\n",
				next, p->line, p->fname);
			break;
		}
	}

	if (tok != 0)
	{
		fprintf(stderr, "error after \"%.*s\" parsing %s, line %d\n", 
			MAXWORDLENGTH, Psz(next), Psz(p->fname), p->line);
		parm->err = 1;
	}
	return parm->err;
}

static int my_lf_accept(int listensock)
/*
** copied from courier/filters/libfilter/libfilter.c
** changed: different return code for shutting down (0 instead of -1)
*/
{
	struct sockaddr_un ssun;
	fd_set fd0;
	int fd;
	int sunlen;

	for (;;)
	{
		FD_ZERO(&fd0);
		FD_SET(0, &fd0);
		FD_SET(listensock, &fd0);
 
		if (select(listensock+1, &fd0, 0, 0, 0) < 0)
		{
			return -1;
		}

		if (FD_ISSET(0, &fd0))
		{
			char buf[16];

			if (read(0, buf, sizeof(buf)) <= 0)
				return 0;    /* 0 is Shutting down (cannot be accepted socket) */
		}

		if (!FD_ISSET(listensock, &fd0))
			continue;

		sunlen = sizeof ssun;
		if ((fd = accept(listensock, (struct sockaddr*)&ssun, &sunlen)) < 0)
			continue;

		fcntl(fd, F_SETFL, 0);  /* Take out of NDELAY mode */
			break;
	}
	
	return fd;
}

int main(int argc, char *argv[])
{
	int rtc = setpgrp(), wait_child = 5;
	av_parm parm;

	if (parm_init(&parm))
		return 1;

	if (rtc)
		fprintf(stderr, "cannot set process group\n");
	else
	{
		struct sigaction act;
		memset(&act, 0, sizeof act);
		sigemptyset(&act.sa_mask);
		
		act.sa_flags = SA_NOCLDSTOP;
		act.sa_handler = child_reaper;
		sigaction(SIGCHLD, &act, NULL);
		
		act.sa_flags = 0;
		act.sa_handler = sig_catcher;
		sigaction(SIGALRM, &act, NULL);
		sigaction(SIGPIPE, &act, NULL);
		sigaction(SIGINT, &act, NULL);
		sigaction(SIGTERM, &act, NULL);
		sigaction(SIGHUP, &act, NULL);
	}

	if (rtc == 0)
		rtc = configure(&parm);

	if (argc > 1 && strcmp(argv[1], "-t") == 0 && rtc == 0)
	{
		int mypipe[2], i;
		if (pipe(mypipe) == 0)
		{
			FILE *fp = fdopen(mypipe[1], "w");
			for (i = 2; i < argc; ++i)
			{
				runchild(mypipe[0], 1, &parm);
#if !defined(NDEBUG)
				fprintf(stderr, "--------- Running %d children from %d\n",
					live_children, (int)getpid());
#endif					
				fprintf(fp, "%s\nthe ids\n\n", argv[i]);
				fflush(fp);
			}
			fclose(fp);
			close(mypipe[0]);
			sleep(1);
		}
		else
			perror("cannot pipe");
	}
	else if (rtc == 0) /* install filter */
	{
		int listensock = lf_init("filters/avfilter-mode",
			ALLFILTERSOCKETDIR "/avfilter",
			ALLFILTERSOCKETDIR "/.avfilter",
			FILTERSOCKETDIR "/avfilter",
			FILTERSOCKETDIR "/.avfilter");

		if (listensock < 0)
		{
			perror("avfilter: lf_init failed");
			return 1;
		}
		
		lf_init_completed(listensock);
		
		for (;;)
		{
			int fd;
			
			if (signal_hangup)
			{
				av_parm p;
				if (parm_init(&p) == 0)
				{
					if (configure(&p) == 0)
					{
						parm_clear(&parm);
						parm = p;
					}
					else
						parm_clear(&p);
				}
				signal_hangup = 0;
			}
			
			if ((fd = my_lf_accept(listensock)) <= 0)
			{
				if (fd < 0) /* select interrupted */
				{
					switch (errno)
					{
						case EAGAIN:
						case EINTR:
							continue;
						default:
							break;
					}
					perror("select");
				}
				break;
			}
			signal_timed_out = signal_break = 0;
			runchild(fd, fd, &parm);
			close(fd); /* -ale: fixed: was missing until 10feb2002 */
		}		
	}

	if (verbose && live_children == 0)
		fprintf(stderr, "avfilter: exiting\n");
	wait_child += live_children;
	while (live_children > 0 && --wait_child > 0)
	{
		int nsec =  3 + live_children/3;
#if !defined(NDEBUG)
		{
			char *debug = getenv("DEBUG_AVFILTER");
			if (!debug)
			{
				fprintf(stderr, "avfilter: waiting for %d child(ren)\n", live_children);
				if (sleep(nsec) != 0)
					continue;
				kill(0, SIGTERM);
			}
		}
#else
		fprintf(stderr, "avfilter: waiting for %d child(ren)\n", live_children);
		if (sleep(nsec) != 0)
			continue;
		kill(0, SIGTERM);
#endif		
	}
	
	parm_clear(&parm);
	return rtc;
}
