/*
** parse.c - written by vesely in milan on may 1997
*/
/*
** Part of MIMEC: see NOTICE.txt
** Copyright (c) 1999, 2000, 2001 Alessandro Vesely
** All rights reserved.
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
#include <errno.h>
#if !defined(NDEBUG) && defined(INCLUDE_DMALLOC)
#define DMALLOC_FUNC_CHECK
#include <dmalloc.h>
#endif

#include "parse.h"

#if defined(TEST_MAIN)
void log_fread(PARSE*p, int size);
#define LOG_FREAD(a,b) log_fread((a), (b))
#else
#define LOG_FREAD(a,b)
#endif /* TEST_MAIN */

const char *last_sys_errmsg()
/* renamed from errmsg to avoid conflict with mysql */
{
	static char msg[256];
	int  err = errno;
	const char *a = strerror(err);
	if (a)
		return a;

	sprintf(msg, "errno=%d", err);
	return msg;
}

const char *Psz(const char *p)
{
	static const char *pnull = "";
	return p && *p ? p : pnull;
}

int stricmp(const char *a, const char *b)
{
	int c, d;
	do c = *a++, d = *b++;
	while (c != 0 && d != 0 && (c == d || (c = tolower(c)) == (d = tolower(d))));

	return c < d ? -1 : c > d;
}

int strincmp(const char *a, const char *b, size_t n)
{
	int c, d;
	while (n)
	{
		c = *a++, d = *b++;
		if (c == 0 || d == 0 || c != d && (c = tolower(c)) != (d = tolower(d)))
			return c < d ? -1 : c > d;
		--n;
	}
	return 0;
}

int is_numeric(const char *p)
{
	int ch;
	while ((ch =*p++) != 0 && isdigit(ch))
		;
	return ch == 0;
}

int parse_invariant(PARSE *p)
{
	return p && p->buf &&
		(p->buffer_size > MAX_READ_SIZE || p->buffer_size > p->total_size) &&
		p->buf <= p->start &&
		p->start <= p->next &&
		p->next <= p->end &&
		p->end < p->endbuf &&
		*p->end == 0 &&
		p->next_token == 0;
		/* 3lug00:last line was: (p->next_token == 0 || ispunct(p->next_token)); */
}

PARSE *parse_init_memory(char const *en, char const *fname, 
	unsigned long total_size, char const *debug)
{
	PARSE *p;

	if (fname == NULL)
		fname = "!memory!";
	if (en == NULL)
	{
		fprintf(stderr, "%s is NULL, no parsing\n", fname);
		return NULL;
	}

	if ((long)total_size == -1L)
		total_size = strlen(en) + 1;

	if ((p = (PARSE*)malloc(sizeof(PARSE))) != NULL)
	{
		char *buf;

		memset(p, 0, sizeof *p);
		assert(MAX_READ_SIZE < PARSE_BUFFER_SIZE);
		p->buffer_size = total_size;

		if ((p->bufarray[0] = buf = (char*)malloc(total_size)) == NULL)
		{
			free(p);
			p = NULL;
		}
		else
		{
			memcpy(buf, en, total_size);
			p->buf = p->start = p->next = buf;
			p->end = buf + total_size - 1;
			p->endbuf = buf + p->buffer_size;
			*p->end = 0;
			p->line = 1;
			p->done = total_size <= 1;
			p->fname = fname;

			if (debug && (p->fp_debug = fopen(debug, "w+")) != NULL)
				fwrite(buf, 1, total_size, p->fp_debug);
		}
	}

	if (p == NULL)
		fprintf(stderr, "memory fault allocating PARSE\n");

	assert(p == NULL || parse_invariant(p));
	return p;
}

PARSE *parse_init_file(FILE* fp, char const *fname, unsigned long total_size, char const *debug)
{
	PARSE* p;

	if (fname == NULL)
		fname = "!no name!";
	if (fp == NULL)
	{
		fprintf(stderr, "Cannot parse %s\n", fname);
		return NULL;
	}

	if ((p = (PARSE*) malloc(sizeof(PARSE))) != NULL)
	{
		char *buf;
		memset(p, 0, sizeof *p);
		assert(MAX_READ_SIZE < PARSE_BUFFER_SIZE);
		p->buffer_size = PARSE_BUFFER_SIZE;
		p->total_size = total_size;

		if ((p->bufarray[0] = buf = (char*)malloc(PARSE_BUFFER_SIZE)) == NULL)
		{
			free(p);
			p = NULL;
		}
		else
		{
			size_t size;

			if (debug)
				p->fp_debug = fopen (debug, "w+");

			size = p->buffer_size - 1;
			if (size > p->total_size)
				size = p->total_size;

			if (size > 0)
			{
				size = fread(buf, 1, size, p->fp = fp);
				if (p->fp_debug)
					fwrite(buf, 1, size, p->fp_debug);
				p->total_size -= size;
			}
			if (ferror(p->fp))
			{
				fprintf(stderr, "cannot read %s: %s\n", fname, last_sys_errmsg());
				fclose(p->fp);
				if (p->fp_debug)
					fclose(p->fp_debug);
				free(buf);
				free(p);
				return NULL;
			}

			p->buf = p->start = p->next = buf;
			p->end = buf + size;
			p->endbuf = buf + p->buffer_size;
			*p->end = 0;
			p->line = 1;
			p->done = size == 0;
			p->fname = fname;
		}
	}

	if (p == NULL)
		fprintf(stderr, "memory fault allocating PARSE\n");

	assert(p == NULL || parse_invariant(p));
	return p;
}

PARSE* parse_init(char const *fname)
/*
** init a new parse struct. fname must stay valid until parse_clear is called.
*/
{
	FILE *fp = fname ? fopen(fname, "rb") : NULL;
	PARSE *p;
	if (fp == NULL)
	{
		fprintf(stderr, "cannot read %s: %s\n",
			fname ? fname : "NULL", last_sys_errmsg());
		return NULL;
	}
	p = parse_init_file(fp, fname, (unsigned long)-1, NULL);
	if (p != NULL)
		p->close_file = 1;
	return p;
}

void parse_clear(PARSE *p)
{
	unsigned int i;
	assert(parse_invariant(p));

	for (i = 0; i < sizeof p->bufarray / sizeof p->bufarray[0]; ++i)
	{
		char *buf = p->bufarray[i];
		if (buf)
			free(buf);
		else
			break; /* the rest is NULLs */
	}
	if (p->fp && p->close_file)
		fclose(p->fp);

	if (p->fp_debug)
		fclose(p->fp_debug);

#if !defined(NDEBUG)
	memset(p, -1, sizeof *p);
#endif

	free(p);
}

static void fatalparseerror(PARSE *p, char const *wherefrom, size_t memsize)
{
	fprintf(stderr, "Fatal parse error %s, "
		"buffer %d, string size %d, buffer size %d, "
		"at line %d in %s\n",
			wherefrom, p->bufndx, memsize, p->buffer_size, p->line, p->fname);
}

static char *save_buffer(PARSE *p, char *end)
/*
** save the buffer, placing 0 on end and resuming from next char
** (if *end is a token, must be saved by caller, p->next must be set by caller)
** -ale:3lug00:allocate a new string when needed to save the delimiter
** (number of buffers doubled for this reason, until they'll be dynamic)
*/
{
	char *begin = p->start;
	int ch = *(unsigned char*)end;
	
	assert(parse_invariant(p));
	assert(begin <= end);

	assert(end <= p->next); /* was: assert(end <= p->next && (end < p->next || p->done)); */

	*end = 0;
	if (end < p->next)
	{
		p->start = end + 1;
	}
	else if (p->done || ch == 0 || isspace(ch))
	{
		if (p->next < p->end)
			++p->next;
		p->start = p->next;
	}
	else /* 3lug00: duplicate data and leave p->start alone */
	{
		size_t memsize = end - begin + 1; /* to duplicate, incl. 0 */
		
		if (memsize < (size_t)(p->endbuf - p->end)) /* move to end of buffer */
		{
			memcpy(p->endbuf -= memsize, begin, memsize);
			begin = p->endbuf;
		}
		else
		{
			char *tmp = p->bufarray[p->bufndx];
			if (p->bufndx + 1 >= sizeof p->bufarray / sizeof p->bufarray[0] ||
				(p->bufarray[p->bufndx] = begin = strdup(begin)) == NULL)
			{
				p->bufarray[p->bufndx] = tmp;
				fatalparseerror(p, "save_buffer", memsize);
				return NULL;
			}
			p->bufarray[++p->bufndx] = tmp;
		}
		*end = ch;
		assert(end == p->next);
	}

	assert(parse_invariant(p));
	/* 3lug00:not valid:
	** assert(begin && begin <= p->start && (begin < p->start || p->done)); */

	return begin;
}


static char *save_buffer_nows(PARSE *p, char *end)
{
	char *begin = p->start;

	assert(parse_invariant(p));
	assert(begin <= end);

	while (--end >= begin &&
		isspace(*(unsigned char*)end))
			continue;
	
	return save_buffer(p, end + 1);
}

static int fillbuf_discard(PARSE *p)
/*
** read more. change buffer as needed. return 0 if successful
** assume p->next has reached p->end, even if not explicitely set
*/
{
	size_t size;
	assert(parse_invariant(p));

	size = p->endbuf - p->start - 1;
	if (p->fp == NULL || feof(p->fp))
		size = 0;
	else if (size < MEMMOVE_RATIO && size < p->total_size ) /* start new buffer */
	{
		char *buf;

		if (++p->bufndx >= sizeof p->bufarray / sizeof p->bufarray[0] ||
			(p->buf = buf = p->bufarray[p->bufndx] =
				(char*) malloc(p->buffer_size)) == NULL)
		{
			fatalparseerror(p, "fillbuf_discard", 0);
			return -1;
		}

		p->start = buf;
		p->endbuf = buf + p->buffer_size;
		size = p->buffer_size - 1;
		buf[size] = 0;
	}

	if (size > MAX_READ_SIZE)
		size = MAX_READ_SIZE;
	if (size > p->total_size)
		size = p->total_size;

	if (size > 0)
	{
		size = fread(p->start, 1, size, p->fp);
		if (p->fp_debug)
			fwrite(p->start, 1, size, p->fp_debug);
		p->total_size -= size;
	}
	p->next = p->start;
	p->end = p->start + size;
	p->done = size == 0;
	*p->end = 0;

	assert(parse_invariant(p));
	LOG_FREAD(p, size);
	return p->fp && ferror(p->fp);
}


static char *fillbuf_keep(PARSE *p, char *out)
/*
** read more. change buffer as needed. return p->next, possibly moved
** assume p->next has reached p->end, even if not explicitely set
*/
{
	size_t size;
	assert(parse_invariant(p));
	assert(out == NULL || out >= p->start && out <= p->end);

	if (out == NULL)
		out = p->end;

	size = p->endbuf - out - 1; /* available */

	if (p->fp == NULL || feof(p->fp))
		size = 0;
	else if (size < MEMMOVE_RATIO && size < p->total_size) /* start new buffer */
	{
		char *buf;
		size_t memsize = out - p->start; /* to be saved */

		if (memsize >= p->buffer_size - MEMMOVE_RATIO ||
			++p->bufndx >= sizeof p->bufarray / sizeof p->bufarray[0] ||
			(p->buf = buf = p->bufarray[p->bufndx] =
				(char*) malloc(p->buffer_size)) == NULL)
		{
			fatalparseerror(p, "fillbuf_keep", memsize);
			return NULL;
		}

		memcpy(buf, p->start, memsize);
		out = buf + memsize;
		p->endbuf = buf + p->buffer_size;
		buf[p->buffer_size - 1] = 0;
		p->start = buf;
		size = p->buffer_size - memsize - 1;
	}

	if (size > MAX_READ_SIZE)
		size = MAX_READ_SIZE;
	if (size > p->total_size)
		size = p->total_size;

	if (size > 0)
	{
		size = fread(out, 1, size, p->fp);
		if (p->fp_debug)
			fwrite(out, 1, size, p->fp_debug);
		p->total_size -= size;
	}

	p->next = out;
	p->end = out + size;
	p->done = size == 0;
	*p->end = 0;

	assert(parse_invariant(p));
	assert(p->next < p->end || p->next == p->end && p->done);
	LOG_FREAD(p, size);

	return p->fp && ferror(p->fp) ? NULL : out;
}

static void get_start_rollback(PARSE *p, char **start)
{
	char *const s = *start;
	
	assert(parse_invariant(p));
	assert(start != NULL);
	
	if (s != NULL && s >= p->buf && s < p->start)
		p->start = s;
	else
		*start = p->start;
		
	assert(parse_invariant(p));
}

#if 0 /* NOT TESTED - situation worsened after 3lug00 - */
static char *fit_in_buffer(PARSE *p, char *saved, int force)
/*
** join the last string that was returned by fillbuf_keep + save_buffer
** taking care of possible buffer changes. return the pointer to the
** previous char (i.e. the terminating 0 of saved) or NULL on error.
** p->next may have been advanced, any saved stuff saved on the same
** buffer after saved will be discarded
** if force is 0, the call will do nothing on a changed buffer
** and will set start to the end of the saved string (unsave).
** note: done must be 0 on entry, unless force is not set.
*/
{
	char *out;
	size_t saved_size = strlen(saved) + 1;

	assert(parse_invariant(p));
	assert(saved != NULL);
	assert(p->done == 0 || force == 0);
	assert(saved_size < p->buffer_size);

	if (saved >= p->buf && saved + saved_size <= p->start)
	/*
	** common case: was (almost) the last saved thing
	** discard [saved+saved_size, p->start)
	*/
	{
		out = saved + saved_size;
		p->start = force ? saved : out;
	}
	else if (force == 0)
		return NULL;
	else
	/*
	** buffer changed: try to move it
	*/
	{
		size_t already = p->start - p->buf;
		size_t new_data = p->end - p->next + 1;
		size_t left_avail = p->next - p->start;
		size_t right_avail = p->endbuf - p->end - 1;
		
		if (saved_size <= left_avail)
		{
			memcpy(p->start, saved, saved_size);
		}
		
		else if (saved_size <= left_avail + right_avail)
		{
			memmove(p->next + right_avail, p->next, new_data);
			p->next += right_avail;
			p->end += right_avail;
			memcpy(p->start, saved, saved_size);
		}
		
		else
		/*
		** cannot move: reallocate
		*/
		{
			size_t needed = saved_size + already + new_data; /* THIS IS THE CORRECT ORDER */
			char *buf;
			
			if (needed < p->buffer_size)
				needed = p->buffer_size;
			if (p->start > p->buf &&
				p->bufndx + 1 >= sizeof p->bufarray / sizeof p->bufarray[0])
					buf = NULL;
			else
				buf = (char*)malloc(needed);
			if (buf == NULL)
			{
				fprintf(stderr,
					"Fatal parse error joining buf %d needed %u in %s line %d\n",
						p->bufndx, needed, p->fname, p->line);
				return -1;
			}

			memcpy(buf, p->buf, already);
			memcpy(buf + already, saved, saved_size);
			memcpy(buf + already + saved_size, p->next, new_data);

			if (p->start > p->buf)
				++p->bufndx;
			else
				free(p->buf);

			p->buf = p->bufarray[p->bufndx] = buf;
			p->start = buf + already;
			p->next = buf + already + saved_size;
			p->end = buf + already + saved_size + new_data;
			p->endbuf = buf + needed;
			if (needed > p->buffer_size)
				p->buffer_size = needed;
			assert(needed == p->buffer_size);
		}
		out = p->start + saved_size - 1;
	}
		
	assert(parse_invariant(p);
	assert(*out == 0);
	assert(out >= p->start && out < p->next);
	
	return out;
}
#endif

int parse_skip(PARSE *p, char *text)
/*
** skip text if an exact match is found
** return -1 for error, 0 for text skipped ok, 1 for text not skipped
** on eof: the function returns 0 even if text is only matched until
** the end of file.
*/
{
	char *cur, *out;
	int ch, ch_t, count = 0;
	assert(parse_invariant(p));

	cur = p->next;
	out = p->start == cur ? NULL : p->start;

	for (;;)
	{
		while ((ch = *cur++) != 0)
		{
			assert(out == NULL || out + 1 < cur);

			ch_t = *text++;
			count += 1;
			if (ch != ch_t || ch_t == 0)
				if (ch_t != 0) /* revert */
				{
					p->next = cur - count;
					if (out)
					{
						assert(p->start < p->next);
						assert(out == p->start + count - 1);

						memmove(p->next, p->start, count - 1);
					}

					assert(parse_invariant(p));
					return 1;
				}
				else /* skip */
				{
					p->next = cur - 1;

					assert(parse_invariant(p));
					return 0;
				}
			else if (out)
				*out++ = ch;
		}

		if ((cur = fillbuf_keep(p, out)) == NULL)
			return -1;

		out = NULL;

		if (p->done)
		{
			p->next = cur - count;

			assert(parse_invariant(p));
			return 1;
		}
	}
}

static int skip_blanks(PARSE *p)
{
	char *cur;
	int ch;
	assert(parse_invariant(p));

	for (;;)
	{
		cur = p->next;
		while ((ch = *cur++) != 0)
			if (!isspace(ch))
			{
				p->next = cur - 1;

				assert(parse_invariant(p));
				return 0;
			}
			else if (ch == '\n')
				++p->line;

		if (fillbuf_discard(p))
			return -1;

		if (p->done)
			return 0;
	}
}

static int skip_toeol(PARSE *p)
{
	char *cur;
	int ch;
	assert(parse_invariant(p));

	for (;;)
	{
		cur = p->next;
		while ((ch = *cur++) != 0)
			if (ch == '\n')
			{
				p->next = cur;
				++p->line;
				return 0;
			}

		if (fillbuf_discard(p))
			return -1;

		if (p->done)
			return 0;
	}
}

int copy_until(PARSE *p, FILE *fp, int delim)
/* copy to fp until the first occurrence of delim (not included) */
{
	char *cur, *begin;
	int ch;
	size_t size;

	assert(parse_invariant(p));
	assert(fp);
	assert(delim != 0);

	for (;;)
	{
		cur = begin = p->next;
		while ((ch = *cur) != 0 && ch != delim)
		{
			if (ch == '\n')
				++p->line;
			++cur;
		}

		if ((size = cur - begin) > 0 && fwrite(begin, size, 1, fp) != 1)
			return -1;

		if (ch == '\n')
			++p->line;

		if (ch && ch == delim)
		{
			p->next = cur;

			assert(parse_invariant(p));
			return 0;
		}

		if (fillbuf_discard(p))
			return -1;

		if (p->done)
			return 0;
	}
}

int get_token_peek(PARSE *p)
/*
** return -1 for error, 0 for eof, %|?"()~& for corresponding symbol,
** '_' for alphnum
** no char consumed, except whitespace
*/
{
	int ch;
	assert(parse_invariant(p));

	if (p->next_token)
	{
		ch = p->next_token;
		if (ch != '#' || p->no_hash_comment) /* -ale: 28sep99 */
			return ch;
		else if (skip_toeol(p))
			return -1;
	}

	for (;;)
	{
		ch = *p->next;
		if (isspace(ch))
		{
			if (skip_blanks(p))
				return -1;
			ch = *p->next;
		}

		if (ch == '#' && p->no_hash_comment == 0) /* -ale: 28sep99 */
		{
			if (skip_toeol(p))
				return -1;
		}
		else if (ch == 0)
		{
			if (p->done)
				return 0;
			else if (fillbuf_discard(p))
				return -1;
		}
		else
			break;
	}

	if (isalnum(ch) || ch == '_')
		ch = '_';

	assert(parse_invariant(p));
	return ch;
}

int get_token(PARSE *p)
/*
** return -1 for error, 0 for eof, %|?"()~& for corresponding symbol,
** '_' for alphnum (no char consumed in this case)
*/
{
	int ch = get_token_peek(p);

	if (p->next_token)
		p->next_token = 0;

	if (ch != 0 && ch != '_')
		++p->next;
		
	assert(parse_invariant(p));
	return ch;
}

int get_if_delimiter(PARSE *p, char const *delimiters)
/*
** return -1 for error, 0 for eof, the delimiter if found,
** 'A' for alphanum or other punct (no char consumed in this case)
*/
{
	int ch;
	assert(parse_invariant(p));

	if (p->next_token)
	{
		ch = p->next_token;
		if (strchr(delimiters, ch) != NULL)
			p->next_token = 0;
		else
			ch = 'A';
		return ch;
	}

	if (skip_blanks(p))
		return -1;

	ch = *p->next;

	if (ch == 0)
		return 0;
	else if (isalnum(ch) || strchr(delimiters, ch) == NULL || (ch & 0x80) != 0)
		ch = 'A';
	else
		++p->next;

	assert(parse_invariant(p));
	return ch;
}

int token_must_be(PARSE *p, int tok)
{
	int mytok = get_token(p);
	if (mytok == tok && mytok >= 0)
		return get_token(p);

	fprintf(stderr, "Invalid token %c: expecting %c in %s at line %d\n",
		mytok, tok, Psz(p->fname), p->line);
	return -1;
}

char *get_next(PARSE *p)
/*
** return pointer to next word, NULL for error (e.g. word too long)
** word is tolowered and 0 terminated
*/
{
	char *cur, *out = p->word, *rtc = out, *end = p->word + sizeof p->word;
	int ch;
	assert(parse_invariant(p));

	if (! isalnum(*(unsigned char*)p->next) && skip_blanks(p))
		return NULL;

	for (;;)
	{
		cur = p->next;
		while ((ch = *(unsigned char*)cur++) != 0)
		{
			if (isalnum(ch) || ch == '_') /* taken */
			{
				*out = tolower(ch);
				if (++out >= end)
					return NULL;
			}
			else
				break;
		}

		if (ch != 0 || p->done)   /* delimiter */
		{
			*out = 0;
			if (out == rtc)
				return NULL;
			if (ch)
				p->next = cur - 1;

			assert(parse_invariant(p));
			return rtc;
		}

		if (fillbuf_discard(p))
			return NULL;
	}
}

char *get_spec(PARSE *p)
/*
** return pointer to next word, NULL for error (e.g. word too long)
** word is case sensitive and may include non-initial '-'
*/
{
	char *cur, *out = p->word, *rtc = out, *end = p->word + sizeof p->word;
	int ch;
	assert(parse_invariant(p));

	if (! isalnum(*(unsigned char*)p->next) && skip_blanks(p))
		return NULL;

	for (;;)
	{
		cur = p->next;
		while ((ch = *(unsigned char*)cur++) != 0)
		{
			if (isalnum(ch) || ch == '_' || ch == '-') /* taken */
			{
				*out = ch;
				if (++out >= end)
					return NULL;
			}
			else
				break;
		}

		if (ch != 0 || p->done)   /* delimiter */
		{
			*out = 0;
			if (out == rtc)
				return NULL;
			if (ch)
				p->next = cur - 1;

			assert(parse_invariant(p));
			return rtc;
		}

		if (fillbuf_discard(p))
			return NULL;
	}
}

char *get_psz_next(PARSE *p)
{
	char *next = get_next(p);
	if (next == NULL)
		next = "";
	return next;
}

char *get_word_until(PARSE *p, int delim)
/*
** return an alloc'd string of text until next whitespace or
** until delimiter. delimiter is consumed silently, if found.
** one whitespace is also consumed (need room for '0')
*/
{
	char *cur, *out;
	int ch;

	assert(parse_invariant(p));

	if (skip_blanks(p))
		return NULL;

	cur = p->next;
	out = cur > p->start ? p->start : NULL;

	for (;;)
	{
		while ((ch = *(unsigned char*)cur++) != 0)
		{
			assert(out == NULL || out + 1 < cur);

			if (ch == delim || isspace(ch))
			{
				if (ch == '\n')
					++p->line;

				p->next_token = 0;
				p->next = cur;

				assert(parse_invariant(p));
				return save_buffer(p, out ? out : cur - 1);
			}

			if (out)
				*out++ = ch;
		}

		if ((cur = fillbuf_keep(p, out)) == NULL)
			return NULL;

		out = NULL;

		if (p->done)
			return save_buffer(p, cur);
	}
}

char *get_toeol(PARSE *p)
/*
** return an alloc'd string of text until end of line or file
** -ale:29jun00: whitespace and comments at the end is discarded
-- CAUTION (currently) empty lines at the current position are discarded --
*/
{
	char *cur, *out;
	int ch;

	assert(parse_invariant(p));

	if (skip_blanks(p))
		return NULL;

	cur = p->next;
	out = cur > p->start ? p->start : NULL;

	for (;;)
	{
		while ((ch = *(unsigned char*)cur++) != 0)
		{
			assert(out == NULL || out + 1 < cur);

			if (ch == '\n' ||
				ch == '#' && !p->no_hash_comment)
			{
				p->next = cur;
				if (ch == '\n')
					++p->line;
				else
					*--p->next = ' '; /* avoid duplicating */
				cur = save_buffer_nows(p, out ? out : cur - 1);
				if (ch == '#')
					skip_toeol(p);

				assert(parse_invariant(p));
				return cur;
			}

			if (out)
				*out++ = ch;
		}

		if ((cur = fillbuf_keep(p, out)) == NULL)
			return NULL;

		out = NULL;

		if (p->done)
			return save_buffer_nows(p, cur);
	}
}

char *get_word_delimited(PARSE *p, char const *deli, int lowcase, int including_deli)
/*
** return an alloc'd string of text (until next non-(alnum or deli)) or NULL
** chars are optionally turned to lowercase, '\\' escapes are not accepted.
** including_deli must be 0 or 1
** deli should include '_' to be consistent with get token (see get_if_delimiter)
*/
{
	char *cur, *out;
	int ch;

	assert(parse_invariant(p));
	assert(deli != NULL);
	assert(including_deli == 0 || including_deli == 1);
	
	if (skip_blanks(p))
		return NULL;

	cur = p->next;
	out = p->start;

	for (;;)
	{
		while ((ch = *(unsigned char*)cur++) != 0)
		{
			assert(out < cur);

			if (isalnum(ch))
			{
				*out++ = lowcase ? tolower(ch) : ch;
			}
			else if (ispunct(ch) &&
				(strchr(deli, ch) != NULL) == including_deli)
			{
				*out++ = ch;
			}
			else /* token is delimiter or newline ends text */
			{
				p->next = cur;
				if (ch == '\n')
					++p->line;
				else if (ispunct(ch))
					--p->next;
				/* else if (ispunct(ch))
					p->next_token = ch; */

				assert(parse_invariant(p));
				return save_buffer(p, out);
			}
		}

		if ((cur = fillbuf_keep(p, out)) == NULL)
			return NULL;

		out = cur;

		if (p->done)
			return save_buffer(p, cur);
	}
}

char *get_word(PARSE *p)
/*
** return an alloc'd string of text (until next non-alnum) or NULL
** chars are turned to lowercase, '\\' escapes are not accepted
*/
{
#if 1 /* replaced by get_word_delimited */
	return get_word_delimited(p, "_", 1, 1);
#else
	char *cur, *out;
	int ch;

	assert(parse_invariant(p));

	if (skip_blanks(p))
		return NULL;

	cur = p->next;
	out = p->start;

	for (;;)
	{
		while ((ch = *cur++) != 0)
		{
			assert(out < cur);

			if (isalnum(ch) || ch == '_')
			{
				*out++ = tolower(ch);
			}
			else /* token is delimiter or newline ends text */
			{
				p->next = cur;
				if (ch == '\n')
					++p->line;
				else if (ispunct(ch))
					p->next_token = ch;

				assert(parse_invariant(p));
				return save_buffer(p, out);
			}
		}

		if ((cur = fillbuf_keep(p, out)) == NULL)
			return NULL;

		out = cur;

		if (p->done)
			return save_buffer(p, cur);
	}
#endif /* replaced by get_word_delimited */
	
}

static char *get_text_var_punct(PARSE *p,
	char const *deli, int const including_deli)
/*
** return an alloc'd string of text (until next token or end of line) or NULL
** white space around text discarded
*/
{
	char *cur, *out;
	int space, ch;
	assert(parse_invariant(p));

	if (skip_blanks(p))
		return NULL;
	space = 0;

	cur = p->next;
	out = p->start == cur ? NULL : p->start;

	for (;;)
	{
		while ((ch = *cur++) != 0)
		{
			assert(out == NULL || out + 1 < cur);

			if (isspace(ch) && ch != '\n')
			{
				if (space == 0) /* first space taken */
				{
					if (out)
						*out++ = ' ';
					else if (ch != ' ')
						*(cur - 1) = ' ';
					space = 1;
				}
				else if (out == NULL) /* subsequent spaces discarded */
					out = cur - 1;
			}
			else if (isalnum(ch) ||
				(strchr(deli, ch) != NULL) == including_deli) /* taken */
			{
				if (out)
					*out++ = ch;
				space = 0;
			}
			else /* token is delimiter or newline ends text */
			{
				p->next = cur;
				if (ch == '\n')
					++p->line;
				else
					--p->next;
				/* else
					p->next_token = ch; */


				assert(parse_invariant(p));
				return save_buffer(p, (out ? out : cur - 1) - space);
			}
		}

		if ((cur = fillbuf_keep(p, out)) == NULL)
			return NULL;

		out = NULL;

		if (p->done)
			return save_buffer(p, cur - space);
	}
}

char *get_text_no_special(PARSE *p, char const *delimiters_punct)
{
	return get_text_var_punct(p, delimiters_punct, 0);
}

char *get_text_including( PARSE *p, char const *alnum_equivalent_punct)
{
	return get_text_var_punct(p, alnum_equivalent_punct, 1);
}

char *get_text(PARSE *p)
/*
** return an alloc'd string of text (until next token or end of line) or NULL
** white space around text discarded
*/
{
#if 1 /* replaced by generic function */
	return get_text_var_punct(p, "-_<>&*{}/\\'`~!@$^:;.", 1);
#else
	char *cur, *out;
	int space, ch;
	assert(parse_invariant(p));

	if (skip_blanks(p))
		return NULL;
	space = 0;

	cur = p->next;
	out = p->start == cur ? NULL : p->start;

	for (;;)
	{
		while ((ch = *cur++) != 0)
		{
			assert(out == NULL || out + 1 < cur);

			if (isspace(ch) && ch != '\n')
			{

				if (space == 0) /* first space taken */
				{
					if (out)
						*out++ = ' ';
					else if (ch != ' ')
						*(cur - 1) = ' ';
					space = 1;
				}
				else if (out == NULL) /* subsequent spaces discarded */
					out = cur - 1;
			}
			else if (isalnum(ch) || strchr("-_<>&*{}/\\'`~!@$^:;.", ch)) /* taken */
			{
				if (out)
					*out++ = ch;
				space = 0;
			}
			else /* token is delimiter or newline ends text */
			{
				p->next = cur;
				if (ch == '\n')
					++p->line;
				else
					p->next_token = ch;

				assert(parse_invariant(p));
				return save_buffer(p, (out ? out : cur - 1) - space);
			}
		}

		if ((cur = fillbuf_keep(p, out)) == NULL)
			return NULL;

		out = NULL;

		if (p->done)
			return save_buffer(p, cur - space);
	}
#endif /* replaced by generic function */
	
}

char *get_until_escaped(PARSE *p, int delim, int do_escape)
/*
** return an alloc'd string of text (until unescaped delim or eof) or NULL
** if do_escape == -1, only the delimiter and '\\'  can be escaped (8jun00)
*/
{
	char *cur, *out;
	int escaped = 0, ch;

	assert(parse_invariant(p));
	assert(ispunct(delim));
	assert(do_escape == 0 || do_escape == 1 || do_escape == -1);

	cur = p->next;
	out = p->start == cur ? NULL : p->start;

	for (;;)
	{
		while ((ch = *cur++) != 0)
		{
			assert(out == NULL || out + 1 < cur);

			if (ch == '\\' && do_escape)
			{
				if (!escaped && do_escape == -1 &&
					*cur != delim && *cur != '\\')
						escaped = 1; /* don't escape *cur */
				if (out == NULL)
					out = cur - 1;
				if (!(escaped = !escaped))
				{
					*out++ = '\\';
					if (out == cur)
						out = NULL;
				}
			}
			else if (escaped)
			{
				switch(ch)
				{
					case 'n':
						ch = '\n';
						break;
					case 'r':
						ch = '\r';
						break;
					case 't':
						ch = '\t';
						break;
					case 'a':
					case '\r':
						ch = 0; /* escaped cr won't print */
						break;
					case '\n':
						++p->line;
						ch = ' '; /* escaped newline prints as space */
						break;
					default: /* escaped delimiter print as-is */
						break;
				}
				escaped = 0;
				if (ch)
				{
					if (out == NULL)
						*(cur - 1) = ch;
					else
						*out++ = ch;
				}
			}
			else /* regular char */
			{
				if (ch == delim)
				{
					/* p->next_token = ch; */
					*--cur = ' '; /* avoid duplicating */
					p->next = cur;
					assert(parse_invariant(p));
					return save_buffer(p, out ? out : cur);
				}
				else if (ch == '\n')
					++p->line;
				else if (ch == '\r')
				{
					if (out == NULL)
						out = cur - 1;
					continue;
				}
				if (out)
					*out++ = ch;
			}
		}	/* end while */

		if ((cur = fillbuf_keep(p, out)) == NULL)
			return NULL;

		out = NULL;

		if (p->done)
			return save_buffer(p, cur);
	}
}

char *get_until(PARSE *p, int delim)
{
	return get_until_escaped(p, delim, 1);
}

char *get_quoted(PARSE *p)
{
	return get_until_escaped(p, '"', 1);
}

char *get_possibly_quoted(PARSE *p)
{
	assert(parse_invariant(p));

	if (skip_blanks(p))
		return NULL;
	if (*p->next == '"')
	{
		int tok = get_token(p);
		assert(tok == '"');

		return get_quoted(p);
	}
	else
		return get_text(p);
}

static int skip_rest_of_parenthesis(PARSE *p, int c_or_d)
/*
** tentative: assume a comment or domain-literal already started and reach
** its end. return non-zero on error (1 soft error, -1 hard error)
**
** comment        =  "(" *(ctext / quoted-pair / comment) ")"
** ctext          =  <any CHAR excluding "(",     ; => may be folded
**                   ")", "\" & CR, & including
**                    linear-white-space>
**
** domain-literal =  "[" *(dtext / quoted-pair) "]"
** dtext          =  <any CHAR excluding "[",     ; => may be folded
**                    "]", "\" & CR, & including
**                    linear-white-space>
**
**
*/
{
	char *cur;
	char *start = p->start;
	int ch, comment = 1, linear_ws = 0, escaped = 0;
	assert(parse_invariant(p));

	for (;;)
	{
		cur = p->next;
		while ((ch = *cur++) != 0)
		{
			if (linear_ws)
			{
				if (ch != ' ' && ch != '\t')
					return 1;
				linear_ws = 0;
			}
			
			if (escaped)
			{
				escaped = 0;
				if (ch == '\n')
					++p->line;				
			}
			else if (ch == '\\')
				++escaped;
			else if (ch == '\"')
			{
				p->next = cur;
				if (get_quoted(p) == NULL)
					return -1;
				
				get_start_rollback(p, &start);
				cur = p->next;
			}
			else if (ch == '\n')
			{
				++p->line;
				linear_ws = 1;
			}
			else if (c_or_d == 'c')
			{
				if (ch == '(')
					++comment;
				else if (ch == ')')
				{
					if (--comment <= 0)
					{
						p->next = cur;
						return 0;
					}
				}
			}
			else if (ch == '[')
				return 1;
			else if (ch == ']')
			{
				p->next = cur;
				return 0;
			}
		}

		if (fillbuf_discard(p) || p->done)
			return -1;
	}
}

static int parse_skip_rfc822_comment(PARSE *p)
/*
** skip rfc822 comment, if any. return non-zero on error
** NOTE: linear_white_space is treated partially
** (may trigger an error in skip_rest_of_parenthesis, not in skip_blanks)
*/
{
	int ch;
	assert(parse_invariant(p));
	
	if (p->next_token)
	{
		ch = p->next_token;
		if (ch == '(')
		{
			p->next_token = 0;
			if (skip_rest_of_parenthesis(p, 'c'))
				return -1;
		}
		return 0;
	}

	for (;;)
	{
		ch = *p->next;
		if (isspace(ch))
		{
			if (skip_blanks(p))
				return -1;
			ch = *p->next;
		}

		if (ch == '(')
		{
			++p->next;
			if (skip_rest_of_parenthesis(p, 'c'))
				return -1;
		}
		else if (ch == 0)
		{
			if (p->done)
				return 0;
			else if (fillbuf_discard(p))
				return -1;
		}
		else
			break;
	}
	
	assert(parse_invariant(p));
	return 0;
}

static int get_quoted_or_delimited(PARSE *p, char **word_ptr, int *deli_ptr,
	char const * const special_delimiter)
/*
** get a word and the following token and return 1; return 0 if did nothing;
** (called internally: this function leaves no next_token, and requires the
** special delimiter string to have the doublequote only at first position)
**
** NOTE: linear_white_space is treated partially
** (a missing space or tab after a line break may trigger an error)
*/
{
	int ch;
	int old_hash;
	char *word = NULL;
	int deli = 0;
	char const * const special_delimiter_noquote = special_delimiter + 1;
	
	assert(parse_invariant(p));
	assert(special_delimiter[0] == '\"');
	assert(strchr(special_delimiter + 1, '\"') == NULL);
	
	old_hash = parse_get_no_hash_comment(p);
	if (old_hash == 0)
		parse_set_no_hash_comment(p, 1);
		
	parse_skip_rfc822_comment(p);
	ch = get_if_delimiter(p, special_delimiter);
	if (ch == '\"')
	{
		/* by rfc822 backslash should be used to escape special chars.
		** however, browsers pass filename="c:\tmp\next\afile"
		** and mailers pass name="half minute is 30\".mov", hence 3rd arg=-1
		** (3rd arg was 0 until 8jun00)
		*/
		word = get_until_escaped(p, '\"', -1);
	}
	else if (ch == 'A')
	{
		word = get_word_delimited(p, special_delimiter, 0, 0);
	}
	else if (ch > 0)
	{
		deli = ch;
	}
	
	if (ch > 0)
	{
		if (word)
		{
			parse_skip_rfc822_comment(p);
			deli = get_if_delimiter(p, special_delimiter_noquote);
		}

		if (word_ptr)
			*word_ptr = word;
		
		if (deli_ptr)
			*deli_ptr = deli;
	}
	
	parse_set_no_hash_comment(p, old_hash);
	assert(parse_invariant(p));
	
	return ch > 0;
}

int get_mime_word(PARSE *p, char **word_ptr, int *deli_ptr)
/*
** get a word and the following token and return 1; return 0 if did nothing; -1 on error
** (called internally: this function leaves no next_token)
**
** NOTE: linear_white_space is treated partially
** (a missing space or tab after a line break may trigger an error)
*/
{
	/* tspecial should be "()<>@,;:\\\"/[]?=": missing ones are not interpreted
	** except backslash in quoted strings and () in comments.
	** note that rfc822 specials is "()<>@,;:\\\".[]"
	*/
	static char const * const tspecial_rfc2045 = "\"()<>@,;:/[]="; /* double quote first! */
	return get_quoted_or_delimited(p, word_ptr, deli_ptr, tspecial_rfc2045);
}

int get_mime_word_relaxed(PARSE *p, char **word_ptr, int *deli_ptr)
/*
** same as get_mime_word, but less constrained; accept "Content-ID: <HisId> "
*/
{
	static char const * const tspecial_relax = "\"();:="; /* double quote first! */
	int rtc = get_quoted_or_delimited(p, word_ptr, deli_ptr, tspecial_relax);
	if (rtc && *word_ptr && **word_ptr == '<')
	{
		unsigned char *s = *word_ptr;
		int ch;
		while ((ch = *++s) != 0)
			continue;
		while (isspace((ch = *--s)))
			continue;
		if (ch == '>')
		{
			*s = 0;
			*word_ptr += 1;
		}
	}
	return rtc;
}

static char const * const rfc822_specials = "\"()<>@,;:.[]="; /* double quote first! */
int get_rfc822_word(PARSE *p, char **word_ptr, int *deli_ptr)
	{ return get_quoted_or_delimited(p, word_ptr, deli_ptr, rfc822_specials); }

int skip_rfc822_domain(PARSE *p, int *deli_ptr, int *count_ptr, int *literal_ptr)
/*
** assume no next_token (after call to get_quoted_or_delimited)
** and return 0 if a valid domain was given, 1 if not, -1 on error
**
** NOTE: linear_white_space is treated partially
** (a missing space or tab after a line break may trigger an error)
*/
{
	int rtc = 0, count = 0, literal = 0, ch = 0;
	char *cur;
	enum expact_enum { start_domain, in_domain, end_domain, done_domain}
		expect = start_domain;
	static char const * specials = "()<>@,;:\\\".[]";
	
	assert(parse_invariant(p));

	for (;;)
	{
		cur = p->next;
		while ((ch = *cur++) != 0)
		{
			if (strchr(specials, ch) != NULL)
			{
				p->next = cur;
				switch (ch)
				{
					case '[':
						if (expect == start_domain)
						{
							rtc = skip_rest_of_parenthesis(p, 'd');
							expect = end_domain;
							literal = 1;
						}
						else
							expect = done_domain;
						break;
						
					case '(':
						rtc = skip_rest_of_parenthesis(p, 'c');
						if (expect == in_domain)
							expect = end_domain;
						break;
					
					case '.':
						if (expect == in_domain || expect == end_domain)
						{
							expect = start_domain;
							break;
						}
						
						/* fall thru if '.' not part of domain */

					default:
						if (expect == start_domain)
							rtc = 1;
						else
							expect = done_domain;
						break;
				}
				
				cur = p->next;
			}
			else if (ch & 0x80 || isspace(ch) || iscntrl(ch))
			{
				if (ch == '\n')
					++p->line;
												
				if (!isspace(ch))
				{
					if (expect == start_domain)
						rtc = 1;
					else
						expect = done_domain;
				}
				else if (expect == in_domain)
					expect = end_domain;
			}
			else if (expect == start_domain)
			{
				expect = in_domain;
				++count;
			}
			else if (expect == end_domain)
				expect = done_domain;
			
			if (rtc)
				return rtc;
			if (expect == done_domain)
				break;
		}

		if (expect == done_domain)
			break;
		
		if (fillbuf_discard(p))
			return -1;
		if (p->done)
		{
			if (expect == start_domain)
				return 1;
			ch = 0;
			break;
		}
	}

	if (deli_ptr)
		*deli_ptr = ch;
	if (literal_ptr)
		*literal_ptr = literal;
	if (count_ptr)
		*count_ptr = count;
	return 0;
}

int skip_rfc822_mailbox(PARSE *p, int *deli_ptr, parse_rfc822_mailbox *desc_ptr)
/*
** break the relevant part of a mailbox (i.e. no group syntax). Data is assumed
** to be unencoded.
** return 0 if successful, return 1 if no valid address found, -1 on error
** use get_mime_word (as removing spaces around '.' or '@' is no use)
**
** NOTE: linear_white_space is treated partially
** (a missing space or tab after a line break may trigger an error)
*/
{
	char *word;
	char *start;
	int deli = 0, rtc, name_found = 0, count, literal;
	parse_rfc822_mailbox desc;
	
	assert(parse_invariant(p));
	
	memset(&desc, 0, sizeof desc);
	start = p->start;
	
	do /* skip initial name or phrase */
	{
		rtc = !get_rfc822_word(p, &word, &deli);
		if (rtc)
			return rtc;
	
		if (word && *word)
			name_found = 1;
		get_start_rollback(p, &start);
	} while (deli == '.');
	
	if (deli != '@' || name_found == 0)
	{
		desc.phrase = name_found > 0;
		while (deli == 'A' || deli == '.') /* skip rest of phrase */
		{
			desc.phrase = 1;
			if ((rtc = !get_rfc822_word(p, NULL, &deli)) != 0)
				return rtc;
			
			get_start_rollback(p, &start);
		}
		
		if (desc.phrase == 0 || deli != '<')
			return 1;

		rtc = !get_rfc822_word(p, &word, &deli); /* skip name or route */
		if (rtc)
			return rtc;

		name_found = word && *word;
		get_start_rollback(p, &start);
			
		while (deli == '@' && name_found == 0) /* skip route "@domain,@domain:" */
		{
			int droute;
			rtc = skip_rfc822_domain(p, &droute, &count, &literal);
			if (rtc)
				return rtc;

			if (droute != ',' && droute != ':')
				return 1;

			++desc.route;
			if (literal)
				desc.literal_route = 1;
			if (count > desc.route_subdomain_max)
				desc.route_subdomain_max = count;
			
			rtc = !get_rfc822_word(p, &word, &deli);
			if (rtc)
				return rtc;

			if (word && *word)
				name_found = 1;
			get_start_rollback(p, &start);
			if (droute == ':')
				break;
		}
		
		while (deli == '.') /* skip rest of name */
		{
			rtc = !get_rfc822_word(p, &word, &deli);
			if (rtc)
				return rtc;

			if (word && *word != 0)
				name_found = 1;
			get_start_rollback(p, &start);
		}
	}

	if (deli != '@' || name_found == 0)
		return 1;
		
	rtc = skip_rfc822_domain(p, &deli, &count, &literal);
	if (rtc)
		return rtc; /* no addr_spec */

	desc.subdomain = count;
	if (literal)
		desc.literal_domain = 1;
	if (desc.phrase)
	{
		if (deli != '>')
			return 1;

		deli = get_if_delimiter(p, rfc822_specials);
	}
	
	if (deli_ptr)
		*deli_ptr = deli;
	if (desc_ptr)
		*desc_ptr = desc;
	
	assert(parse_invariant(p));
	return 0;
}

char *get_urlencoded(PARSE *p)
{
#define HEX_VALUE(ch) (isdigit(ch) ? (ch) - '0' : \
							isupper(ch) ? (ch) - 'A' + 10 : (ch) - 'a' + 10)

	char *cur, *out;
	int unescape = 0, ch, chex1 = 0;
	assert(parse_invariant(p));

	cur = p->next;
	out = p->start;

	for (;;)
	{
		while ((ch = *cur++) != 0)
		{
			assert(out && out < cur);

			switch(unescape)
			{
				case 1:
					chex1 = ch;
					unescape = 2;
					break;

				case 2:
					if (isxdigit(chex1) && isxdigit(ch))
					{
						out -= 2;
						ch = HEX_VALUE(ch) + 16 * HEX_VALUE(chex1);
						if (ch == '\r')
							ch = 0;
					}
					unescape = 0;
					break;

				default:
					if (ch == '+')
						ch = ' ';
					else if (ch == '%')
					{
						unescape = 1;
					}
					/* -ale: bug fix added "@#" in the search string on 4 lug 97 */
					else if (! isalnum(ch) && strchr("$@#_-.^!*(),", ch) == 0)
					{
						p->next = cur;
						/* if (ispunct(ch))
							p->next_token = ch;
						else */
						if (ispunct(ch))
							--p->next;
						else if (ch == '\n')
							++p->line;

						assert(parse_invariant(p));
						return save_buffer(p, out ? out : cur	- 1);
					}
					break;
			}
			if (ch)
				*out++ = ch;
		}	/* end while */

		if ((cur = fillbuf_keep(p, out)) == NULL)
			return NULL;

		out = cur;

		if (p->done)
			return save_buffer(p, cur);
	}
#undef HEX_VALUE
}

#if defined(TEST_MAIN)

void cdispl(char *dest, char *edest, char *src, char *esrc)
{
	char *buf, *pb = src, *pc;
	int dsize, ssize;

	assert(dest && edest && src && esrc);
	assert(dest + 10 < edest);
	assert(src <= esrc);

	dsize = edest - dest;
	memset(dest, 0, dsize);

	buf = pc = (char*) malloc(4*(esrc - src) + 1); /* brute force */
	if (buf == NULL)
		return;

	while (pb < esrc)
	{
		int ch = *pb++;
		if (isalnum(ch) || ispunct(ch) || ch == ' ')
		{
			*pc++ = ch;
			if (ch == '\\')
				*pc++ = ch;
		}
		else
		{
			*pc++ = '\\';
			switch (ch)
			{
				case '\r': *pc++ = 'r'; break;
				case '\n': *pc++ = 'n'; break;
				case '\t': *pc++ = 't'; break;
				case '\a': *pc++ = 'a'; break;
				default:
					pc += sprintf(pc, "x%02X", ch & 0xff);
					break;
			}
		}
	}

	ssize = pc - buf;
	if (ssize < dsize)
		memcpy(dest, buf, ssize);
	else
	{
		dsize /= 2;
		memcpy(dest, buf, dsize);
		memcpy(edest - dsize - 1, pc - dsize, dsize);
		memcpy(dest + dsize - 1, "...", 3);
	}

	free(buf);
}

int print_parse(PARSE *p)
{
	static const char* desc[4] = { "buf", "start", "next", "end"};
	char *ptr[5];
	char content[80];
	char *nextptr;
	unsigned int i;


	if (p == NULL)
		return printf("p = NULL\n");

	if (! parse_invariant(p))
		return -1;

	printf("Buffers:");
	for (i = 0; i < sizeof p->bufarray / sizeof p->bufarray[0]; ++i)
		if (p->bufarray[i] != NULL)
		{
			static const char*const desc[4] = { "", "", "-->", "<--"};
			int z = i == p->bufndx ? 2 : 0;
			printf(" %s%d%s", desc[z], i, desc[z+1]);
		}
	printf("\n");

	printf("buffer %d, size = %d, line %d of %s, ",
		p->bufndx, p->buffer_size, p->line, p->fname);
	if (p->word[0] || p->next_token)
		printf("next_token = %c (%d = x%x), word = %.*s\n",
			p->next_token, p->next_token, p->next_token, MAXWORDLENGTH, p->word);
	else printf("\n");

	ptr[0] = p->buf;
	ptr[1] = p->start;
	ptr[2] = p->next;
	ptr[3] = p->end;
	ptr[4] = p->endbuf;

	for (i = 0; i < 4; ++i)
	{
		nextptr = ptr[i+1];
		cdispl(content, content + sizeof content, ptr[i], nextptr);
		printf("%8p (%4d) %-7s %s\n",
			ptr[i], nextptr - ptr[i], desc[i], content);
	}
	printf("remaining %d until endbuf, done = %d, feof = %s, ferror = %s\n\n",
		p->endbuf - p->end, p->done,
		p->fp ? feof(p->fp) ? "y" : "n" : "n/a",
		p->fp ? ferror(p->fp) ? "y" : "n" : "n/a");
	return 0;
}

void log_fread(PARSE *p, int size)
{
	printf("read %d bytes\n", size);
	if (size > 0)
		print_parse(p);
}



int main(int argc, char *argv[])
{
	PARSE *p = NULL;
	char *fname = NULL;
	int i, run = 1;
	int ch, r;
	char cmd[1024], *s;

	for (i = 1; run; ++i)
	{
		s = cmd;
		if (i < argc)
		{
			strcpy(cmd, argv[i]);
		}
		else
		{
			printf("%d>", i);
			fflush(stdout);
			gets(cmd);
			while (*s && isspace(*(unsigned char*)s))
				++s;
		}

		if (p == NULL)
		{
			if (strcmp(s, "?") == 0)
			{
				printf(
					"to start parsing a file enter the filename\n"
					"to parse memory enter an open quote (\") and the text to parse\n"
					"to exit enter 'x'\n");
			}
			else if (stricmp(s, "x") == 0)
				run = 0;
			else if (*s == '"')
			{
				p = parse_init_memory(++s, "test string", -1, NULL);
				if (p == NULL)
					printf("parse_init_memory failed\n");
			}
			else if (*s)
			{
				p = parse_init(fname = strdup(cmd));
				if (p == NULL)
				{
					printf("parse_init failed init with %s\n", Psz(fname));
					free(fname);
					fname = NULL;
				}
			}
		}
		else
		{
			char *(*fn)(PARSE*) = NULL;
			while ((ch = *s++))
			{
				if (ch == '?')
					printf(
						"s   parse_skip(rest)  k   get_token()\n"
						"n   get_next()        w   get_word()\n"
						"t   get_text()        a   get_any_word()\n"
						"q   get_quoted()      p   get_possibly_quoted()\n"
						"c   get_urlencoded()  l   get_toeol()\n"
						"\"<delim>\" [li]        w   get_word_delimited(delim, l?, i?)\n"
						"\"<delim>\" [i]         t   get_text_(i?including:no_special)\n"
						"\"<delim>\"             g   get_if_delimiter()\n"
						"u%%  get_until '%%'     uc%% copy until '%%'\n"
						"uw%% get_word_until '%%'\n"
						"mm  get_mime_word()   mb  discard mailbox\n"
						"mr  get_mime_word_rlx m8  get_rfc822_word()\n"
						"?   this help         +   print PARSE struct\n"
						"x  exit               X   close\n");
				else if (ch == 'x' || ch == 'X')
				{
					parse_clear(p);
					p = NULL;
					if (fname)
					{
						free(fname);
						fname = NULL;
					}
					if (ch == 'x')
						run = 0;
					break;
				}
				else if (p != NULL)
				{
					switch(ch)
					{
						case 's':
							r = parse_skip(p, s);
							printf("s(%s) %d\n", s, r);
							*s = 0;
							break;
						case 'k':
							r = get_token(p);
							printf("k %c (%d = x%x)\n", r, r, r);
							break;
						case 'n': fn = get_next; break;
						case 'w': fn = get_word; break;
						case 'a': fn = get_any_word; break;
						case 't': fn = get_text; break;
						case 'q': fn = get_quoted; break;
						case 'c': fn = get_urlencoded; break;
						case 'p': fn = get_possibly_quoted; break;
						case 'l': fn = get_toeol; break;
						case '+':
							print_parse(p);
							break;
						case 'u':
							if ((ch = *s))
							{
								int delim = *++s;
								if (delim == 0 && (ch == 'w' || ch == 'c'))
									printf("u%c must be followed by delimiter\n", ch);
								else
								{
									if (ch == 'c')
									{
										printf("\n====output begin for copy_until(%c)===\n", delim);
										r = copy_until(p, stdout, delim);
										printf("\n====output ends for copy_until(%c)=%d\n",
											delim, r);
										++s;
									}
									else if (ch == 'w')
									{
										char *q = get_word_until(p, delim);
										char content[20];
										if (q)
										{
											cdispl(content, content + sizeof content,
												q, q + strlen(q));
											printf("uw %c %s\n", delim, content);
										}
										else
											printf("uw %c is NULL\n", delim);
										++s;
									}
									else
									{
										char *q = get_until(p, ch);
										char content[20];
										if (q)
										{
											cdispl(content, content + sizeof content,
												q, q + strlen(q));
											printf("u %c %s\n", delim, content);
										}
										else
											printf("u %c is NULL\n", delim);
									}
								}
							}
							else
								printf("u must be followed by [c|w] delim\n");
							break;
						case 'm':
						{
							int mydeli = 0, rtc;
							char *unchanged = "u";
							char *myword = unchanged;
							parse_rfc822_mailbox mydesc;
							
							parse_set_no_hash_comment(p, 1);
							ch = *s++;
							switch (ch)
							{
								case 'm':
									rtc = get_mime_word(p, &myword, &mydeli);
									break;
								case 'r':
									rtc = get_mime_word_relaxed(p, &myword, &mydeli);
									break;
								case '8':
									rtc = get_rfc822_word(p, &myword, &mydeli);
									break;
								case 'b':
									rtc = skip_rfc822_mailbox(p, &mydeli, &mydesc);
									break;
								default:
									printf("invalid m%c: m must be followed by [m|8|b]\n", ch);
									ch = rtc = 0;
									break;
							}
							if (ch)
							{
								printf ("m%c rtc=%d, deli=%c [%d]", ch, rtc,
									isprint(mydeli) ? mydeli : ' ', mydeli);
								if (ch != 'b')
									if (myword == unchanged)
										printf(", word UNCHANGED");
									else if (myword)
										printf(", word = %p: -->%s<--, length=%d",
											myword, myword, strlen(myword));
									else
										printf(", word = NULL");
								else if (rtc == 0)
									printf(" desc:%s%s route-length=%d "
											"max-route-subdomain=%d%s sub-domain-count=%d",
										mydesc.phrase ? " phrase" : "",
										mydesc.literal_route ? " literal-in-route" : "",
										mydesc.route,
										mydesc.route_subdomain_max,
										mydesc.literal_domain ? " literal-in-domain" : "",
										mydesc.subdomain);
								printf("\n");
							}
							parse_set_no_hash_comment(p, 0);
							break;
						}
						case '\"':
						{
							char delim[64], *pd = delim, *epd = delim + sizeof delim - 1;
							char content[32], rtc[4], *q = rtc;
							int i_switch = 0, l_switch = 0, d_command = '?';
							
							while ((ch = *s++) != 0 && pd < epd && ch != '\"')
								*pd++ = ch;
							*pd = 0;
							if (ch == '\"')
								while ((ch = *s++) != 0)
								{
									if (ch == 'i' && i_switch == 0)
										i_switch = 1;
									else if (ch == 'l' && l_switch == 0)
										l_switch = 1;
									else if (strchr("wtg", ch) != NULL)
									{
										d_command = ch;
										break;
									}
									else if (!isspace(ch))
									{
										--s;
										break;
									}
								}
							
							switch (d_command)
							{
								case 'w':
									q = get_word_delimited(p, delim, l_switch, i_switch);
									break;
								case 't':
									if (i_switch)
										q = get_text_including(p, delim);
									else
										q = get_text_no_special(p, delim);
									break;
								case 'g':
									rtc[0] = get_if_delimiter(p, delim);
									rtc[1] = 0;
									break;
								default:
									printf("invalid syntax\n");
									rtc[0] = 0;
									break;
							}
							if (q)
								cdispl(content, content + sizeof content, q, q + strlen(q));
							else
								strcpy(content, "(null)");
							printf("\"%s\"%s%s%c %s\n", delim,
								l_switch ? "l" : "", i_switch ? "i" : "", d_command, content);
							break;
						}
						default:
							printf("invalid %c\n", ch);
							break;
					}
					if (fn)
					{
						char *q = (*fn)(p);
						char content[20];
						if (q)
						{
							cdispl(content, content + sizeof content,
								q, q + strlen(q));
							printf("%c %s\n", ch, content);
						}
						else
							printf("%c is NULL\n", ch);
						fn = NULL;
					}
				}
			}
		}
	}
	if (p)
	{
		parse_clear(p);
		free(fname);
	}
	return 0;
}
#endif /* TEST_MAIN */

